'use strict'

const fs = require('fs')
const fse = require('fs-extra')
const path = require('path')
const {
  log,
  inquirer,
  spinner,
  Package,
  sleep,
  exec,
  validateNpmPackageName,
  validateNpmPackageVersion,
  ejs
} = require('@jqq-cli-dev/util')
const getProjectTemplate = require('./getProjectTemplate')
const COMPONENT_FILE = '.componentrc'
const TYPE_COMPONENT = 'component'
const TYPE_PROJECT = 'project'
const TEMPLATE_TYPE_NORMAL = 'normal'
const TEMPLATE_TYPE_CUSTOM = 'custom'

const DEFAULT_TYPE = TYPE_PROJECT

module.exports = init

async function init (options) {
  try {
    // 初始化 targetPath
    const targetPath = process.cwd()
    if (!options.targetPath) {
      options.targetPath = targetPath
    }
    log.verbose('init', options)

    // 完成项目初始化准备和检验工作
    const result = await prepare(options)
    if (!result) {
      log.info('缺少相关信息，创建项目终止')
      return
    }

    // 获取项目模板列表
    const { templateList, project } = result

    // 缓存项目模板文件
    const template = await downloadTemplate(templateList, options)
    // log.verbose('template', template)

    if (template.type === TEMPLATE_TYPE_NORMAL) {
      // 安装默认模板
      await installTemplate(template, project, options)
    } else if (template.type === TEMPLATE_TYPE_CUSTOM) {
      // 安装自定义模板
      await installCustomTemplate(template, project, options)
    } else {
      throw new Error('未知的模板类型！')
    }
  } catch (e) {
    if (options.debug) {
      // 打印错误堆栈信息
      log.error('Error', e.stack)
    } else {
      log.error('Error', e.message)
    }
  } finally {
    process.exit(0)
  }
}

async function prepare (options) {
  let fileList = fs.readdirSync(options.targetPath)

  // 目录非空的处理
  fileList = fileList.filter((file) => !['node_modules', '.git', '.DS_Store'].includes(file))

  let continueWhenDirNotEmpty = true
  if (fileList && fileList.length > 0) {
    continueWhenDirNotEmpty = await inquirer({
      type: 'confirm',
      message: '当前文件夹不为空，是否继续创建项目',
      defaultValue: false
    })

    // 不继续则退出选择
    if (!continueWhenDirNotEmpty) {
      return
    }

    // 确认是否清空目录
    const targetDir = options.targetPath
    const confirmEmptyDir = await inquirer({
      type: 'confirm',
      message: '是否确认清空当前目录下的文件',
      defaultValue: false
    })

    if (confirmEmptyDir) {
      fse.emptyDirSync(targetDir)
    }
  }

  // 获取用户选择
  const initType = await getInitType()

  // 获取项目模板
  try {
    let templateList = await getProjectTemplate()
    log.verbose('templateList', templateList)
    if (!templateList || templateList.length === 0) {
      throw new Error('项目模板列表获取失败')
    }

    // 获取项目信息
    let projectName = ''
    let className = ''
    do {
      // 输入项目或组件名称
      projectName = await getProjectName(initType)
      const isValidate = validateNpmPackageName(projectName)
      if (!isValidate) {
        projectName = ''
        className = ''
        log.error('项目名称不符合规范, 请重新输入')
      }

      log.verbose('name', projectName)
      log.verbose('className', className)
    } while (!projectName)

    let version = '1.0.0'

    do {
      // 获取项目版本
      version = await getProjectVersion(version, initType)
      const isValidate = validateNpmPackageVersion(version)
      if (!isValidate) {
        version = ''
        log.error('版本号不符合规范(major >= 1), 请重新输入')
      }
      log.verbose('version', version)
    } while (!version)

    // 选择项目模板，返回模板和项目信息
    if (initType === TYPE_PROJECT) {
      templateList = templateList.filter((item) =>
        item.tag.includes('project')
      )
      return {
        templateList,
        project: {
          name: projectName,
          className,
          version
        }
      }
    } else {
      // 选择组件模板，返回模板和组件信息
      templateList = templateList.filter((item) => item.tag.includes('component'))
      let description = ''
      while (!description) {
        description = await getComponentDescription()
        log.verbose('description', description)
      }
      return {
        templateList,
        project: {
          name: projectName,
          className,
          version,
          description
        }
      }
    }
  } catch (err) {
    log.verbose('err', err)
  }
}

// 获取用户选择
function getInitType () {
  return inquirer({
    type: 'list',
    choices: [{
      name: '项目',
      value: TYPE_PROJECT
    }, {
      name: '组件',
      value: TYPE_COMPONENT
    }],
    message: '请选择初始化类型',
    defaultValue: DEFAULT_TYPE
  })
}

// 获取项目或组件命名
function getProjectName (initType) {
  return inquirer({
    type: 'string',
    message: initType === TYPE_PROJECT ? '请输入项目名称' : '请输入组件名称',
    defaultValue: ''
  })
}

// 获取项目版本信息
function getProjectVersion (defaultVersion, initType) {
  return inquirer({
    type: 'string',
    message: initType === TYPE_PROJECT ? '请输入项目版本号' : '请输入组件版本号',
    defaultVersion
  })
}

// 获取组件信息
function getComponentDescription () {
  return inquirer({
    type: 'string',
    message: '请输入组件的描述信息',
    defaultValue: ''
  })
}

// 下载项目模板
async function downloadTemplate (templateList, options) {
  // 用户交互选择
  const templateName = await inquirer({
    choices: createTemplateChoice(templateList),
    message: '请选择项目模板'
  })

  log.verbose('templateName', templateName)

  const selectedTemplate = templateList.find((item) => (item.npmName === templateName))

  const { cliHome } = options
  const targetPath = path.resolve(cliHome, 'template')
  const storePath = path.resolve(cliHome, 'template', 'node_modules')

  // 基于模板生成 Package 对象
  const tempaltePkg = new Package({
    targetPath,
    storePath,
    packageName: selectedTemplate.npmName,
    packageVersion: selectedTemplate.version
  })

  // 如果模板不存在则进行下载
  if (!await tempaltePkg.exists()) {
    const spinnerIns = spinner()
    spinnerIns.start('正在下载模板...' + '\n')
    await sleep(1000)
    await tempaltePkg.install()
    spinnerIns.succeed('下载模板成功')
  } else {
    log.notice('模板已存在', `${selectedTemplate.npmName}@${selectedTemplate.version}`)
    log.notice('模板路径', `${targetPath}`)
    const spinnerIns = spinner()
    spinnerIns.start('开始更新模板...' + '\n')
    await sleep(1500)
    await tempaltePkg.update()
    spinnerIns.succeed('更新模板成功')
  }

  // 获取模板安装路径
  const templateSourcePath = tempaltePkg.npmFilePath

  // node_modules/_@jqq-cli-template_sub-application@1.1.5@@jqq-cli-template/sub-application
  const tempaltePath = path.resolve(templateSourcePath, 'template')
  log.verbose('template path', tempaltePath)
  if (!fse.pathExistsSync(tempaltePath)) {
    throw new Error(`[${templateName}]项目模板不存在！`)
  }

  const template = {
    ...selectedTemplate,
    path: tempaltePath,
    sourcePath: templateSourcePath
  }

  // 安装模板成功后的返回相关信息
  return template
}

function createTemplateChoice (list) {
  return list.map((item) => ({
    value: item.npmName,
    name: item.name
  }))
}

// 安装默认模板
async function installTemplate (template, ejsData, options) {
  log.verbose('installTemplate template', template)
  log.verbose('installTemplate ejsData', ejsData)
  log.verbose('installTemplate options', options)

  // 安装（npm install) 模板
  const spinnerIns = spinner()
  spinnerIns.start('正在安装模板...' + '\n')
  await sleep(1000)

  // npm 模板默认存放的位置
  const sourceDir = template.path

  // npm 模板转换成自己指定的安装路径
  const targetDir = options.targetPath

  fse.ensureDirSync(sourceDir)
  fse.ensureDirSync(targetDir)
  fse.copySync(sourceDir, targetDir)
  spinnerIns.succeed('安装模板成功')

  // ejs 模板渲染
  const ejsIgnoreFiles = [
    '**/node_modules/**',
    '**/.git/**',
    '**/.vscode/**',
    '**/.DS_Store'
  ]

  if (template.ignore) {
    ejsIgnoreFiles.push(...template.ignore)
  }
  await ejs(targetDir, ejsData, {
    ignore: ejsIgnoreFiles
  })

  // 如果是组件，进行特殊处理 TODO:
  await createComponentFile(template, ejsData, targetDir)

  // 安装依赖文件
  spinnerIns.info('开始安装依赖...')

  // 安装模板中的依赖
  try {
    await npminstall(targetDir)
    spinnerIns.succeed('依赖安装成功')

    // 启动工程
    if (template.startCommand) {
      log.notice('开始执行启动命令')
      const startCommand = template.startCommand.split(' ')
      if (['npm', 'cnpm'].includes(startCommand[0])) {
        await execStartCommand(targetDir, startCommand)
      }
    }
  } catch (err) {
    log.verbose('installTemplate', err)
  }
}

// 安装自定义模板
async function installCustomTemplate (template, ejsData, options) {
  const pkgPath = path.resolve(template.sourcePath, 'package.json')
  const pkg = fse.readJSONSync(pkgPath)
  const rootFile = path.resolve(template.sourcePath, pkg.main)
  if (!fs.existsSync(rootFile)) {
    throw new Error('入口文件不存在！')
  }
  log.notice('开始执行自定义模板')
  const targetPath = options.targetPath
  await execCustomTemplate(rootFile, {
    targetPath,
    data: ejsData,
    template
  })
  log.success('自定义模板执行成功')
}

function execCustomTemplate (rootFile, options) {
  // rootFile 命令, options 命令参数
  const code = `require('${rootFile}')(${JSON.stringify(options)})`
  return new Promise((resolve, reject) => {
    const p = exec('node', ['-e', code], { stdio: 'inherit' })
    p.on('error', e => {
      reject(e)
    })

    p.on('exit', c => {
      resolve(c)
    })
  })
}

// 如果是组件项目，写入相关信息到 .componentrc 中
async function createComponentFile (template, data, dir) {
  if (template.tag.includes(TYPE_COMPONENT)) {
    const componentData = {
      ...data,
      buildPath: template.buildPath,
      examplePath: template.examplePath,
      npmName: template.npmName,
      npmVersion: template.version
    }
    const componentFile = path.resolve(dir, COMPONENT_FILE)
    fs.writeFileSync(componentFile, JSON.stringify(componentData))
  }
}

async function npminstall (targetPath) {
  return new Promise((resolve, reject) => {
    const p = exec('npm', ['install', '--registry=https://registry.npm.taobao.org'], { stdio: 'inherit', cwd: targetPath })
    p.on('error', e => {
      log.verbose('npminstall error', e)
      reject(e)
    })
    p.on('exit', c => {
      log.verbose('npminstall exit', c)
      resolve(c)
    })
  })
}

async function execStartCommand (targetPath, startCommand) {
  return new Promise((resolve, reject) => {
    const p = exec(startCommand[0], startCommand.slice(1), { stdio: 'inherit', cwd: targetPath })
    p.on('error', e => {
      reject(e)
    })
    p.on('exit', c => {
      resolve(c)
    })
  })
}
